#ifndef _CHRONOPTICS_TOF_CAMERA_H_
#define _CHRONOPTICS_TOF_CAMERA_H_

#include <chronoptics/tof/stream.h>
#include <chronoptics/tof/processing_config.h>
#include <chronoptics/tof/data.h>
#include <chronoptics/tof/camera_config.h>
#include <chronoptics/tof/calibration.h>

#ifdef __cplusplus
extern "C" {
#endif

/** The main interface to the depth cameras
 */
typedef struct tof_camera* tof_camera_t;

typedef struct tof_camera const* tof_camera_ct;

/** Destruct tof_camera */
TOF_EXPORT void tof_camera_delete(tof_camera_t ptr);

/** Start streaming of the camera
 * @param ptr Pointer to class
 * @param error Pointer to error
 */
TOF_EXPORT void tof_camera_start(tof_camera_t ptr, tof_error_t *error);

/** Stop streaming of the camera
 * @param ptr Pointer to class
 * @param error Pointer to error
 */
TOF_EXPORT void tof_camera_stop(tof_camera_t ptr, tof_error_t *error);

/** Check whether the camera is streaming
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Is streaming
 */
TOF_EXPORT bool tof_camera_is_streaming(tof_camera_ct ptr, tof_error_t *error);

/** Check whether the camera is still connected
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Is connected
 */
TOF_EXPORT bool tof_camera_is_connected(tof_camera_ct ptr, tof_error_t *error);

/** Get the different output streams the camera can provide
 * @param ptr Pointer to class
 * @param streams The streams
 * @param capacity The amount of data the above pointer can hold
 * @param error Pointer to error
 * @return Number of pointers filled or complete array size
 */
TOF_EXPORT size_t tof_camera_get_stream_list(tof_camera_ct ptr, tof_stream_t* streams, size_t capacity, tof_error_t *error);

/** the processing config of the camera
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The processing config
 */
TOF_EXPORT tof_processing_config_t tof_camera_get_process_config(tof_camera_ct ptr, tof_error_t *error);

/** Check if the camera has frames available
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Has frames available
 */
TOF_EXPORT bool tof_camera_has_frames(tof_camera_ct ptr, tof_error_t *error);

/** Get data frames from camera
 * @param ptr Pointer to class
 * @param frames Data frames
 * @param capacity The amount of data the above pointer can hold
 * @param error Pointer to error
 * @return Number of pointers filled or complete array size
 */
TOF_EXPORT size_t tof_camera_get_frames(tof_camera_t ptr, tof_data_t* frames, size_t capacity, tof_error_t *error);

/** Clear the circular buffer used to store frames
 * @param ptr Pointer to class
 * @param error Pointer to error
 */
TOF_EXPORT void tof_camera_clear_buffer(tof_camera_t ptr, tof_error_t *error);

/** Get the circular buffer size
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The maximum number of sets of frames that is stored before the oldest
 * is overwritten
 */
TOF_EXPORT size_t tof_camera_get_buffer_size(tof_camera_ct ptr, tof_error_t *error);

/** Set the circular buffer size
 * @param ptr Pointer to class
 * @param size The maximum number of sets of frames that is stored before the
 * oldest is overwritten
 * @param error Pointer to error
 */
TOF_EXPORT void tof_camera_set_buffer_size(tof_camera_t ptr, size_t size, tof_error_t *error);

/** Get the currently active camera config
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The camera config
 */
TOF_EXPORT tof_camera_config_t tof_camera_get_camera_config(tof_camera_ct ptr, tof_error_t *error);

/** Get the serial number of the current camera
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The serial
 */
TOF_EXPORT const char* tof_camera_get_serial(tof_camera_ct ptr, tof_error_t *error);

/** Get calibration from the camera
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The calibration
 */
TOF_EXPORT tof_calibration_t tof_camera_get_calibration(tof_camera_ct ptr, tof_error_t *error);

/** Get the amount of user pointers that can be stored
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The capacity
 */
TOF_EXPORT size_t tof_camera_get_user_pointer_capacity(tof_camera_ct ptr, tof_error_t *error);

/** Set the amount of user pointers that can be stored
 * @param ptr Pointer to class
 * @param capacity The capacity
 * @param error Pointer to error
 */
TOF_EXPORT void tof_camera_set_user_pointer_capacity(tof_camera_t ptr, size_t capacity, tof_error_t *error);

/** Add a pointer that will be filled with the specified data type. This method
 * allows you to fill data straight into your own data type without any
 * additional copying.
 * @param ptr Pointer to class
 * @param pointer The pointer to the data to write into
 * @param capacity The amount of data the pointer can hold
 * @param callback Callback that will be called when pointer is no longer in use
 * @param callback_user_data User data that will be passed back when the function pointer is called
 * @param frame_type The frame type that will be filled into the pointer
 * @param error Pointer to error
 */
TOF_EXPORT void tof_camera_add_user_pointer(tof_camera_t ptr, uint8_t* pointer, size_t capacity, tof_user_pointer_destructed_t callback, void* callback_user_data, enum tof_frame_type frame_type, tof_error_t *error);

/** Clear all user pointers in memory
 * @param ptr Pointer to class
 * @param error Pointer to error
 */
TOF_EXPORT void tof_camera_clear_user_pointers(tof_camera_t ptr, tof_error_t *error);

/** Get the currently set configuration index
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return config index
 */
TOF_EXPORT uint32_t tof_camera_config_index(tof_camera_ct ptr, tof_error_t *error);

/** Switch to a different config in real time, depending on configuration and
 * camera, this can take 100-300ms
 * @param ptr Pointer to class
 * @param config_index Config index to switch to
 * @param error Pointer to error
 */
TOF_EXPORT void tof_camera_switch_config(tof_camera_t ptr, uint32_t config_index, tof_error_t *error);

/** Get the frame types that this camera can output
 * @param ptr Pointer to class
 * @param frame_types Possible frame types
 * @param capacity The amount of data the above pointer can hold
 * @param error Pointer to error
 * @return Number of pointers filled or complete array size
 */
TOF_EXPORT size_t tof_camera_possible_frame_types(tof_camera_ct ptr, enum tof_frame_type* frame_types, size_t capacity, tof_error_t *error);

/** Whether output frame types have been set
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Output frame types set
 */
TOF_EXPORT bool tof_camera_output_frame_types_set(tof_camera_t ptr, tof_error_t *error);

/** Get the frame types the camera should output
 * @param ptr Pointer to class
 * @param output_frame_types Output frame types
 * @param capacity The amount of data the above pointer can hold
 * @param error Pointer to error
 * @return Number of pointers filled or complete array size
 */
TOF_EXPORT size_t tof_camera_get_output_frame_types(tof_camera_ct ptr, enum tof_frame_type* output_frame_types, size_t capacity, tof_error_t *error);

/** Set the frame types the camera should output
 * @param ptr Pointer to class
 * @param output_frame_types Output frame types
 * @param output_frame_types_size The size of the above array
 * @param error Pointer to error
 */
TOF_EXPORT void tof_camera_set_output_frame_types(tof_camera_t ptr, const enum tof_frame_type* output_frame_types, size_t output_frame_types_size, tof_error_t *error);

/** Check if multiple configurations are set. Switch config can only be used
 * when this is the case.
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Is set
 */
TOF_EXPORT bool tof_camera_has_multiple_configurations(tof_camera_ct ptr, tof_error_t *error);

/** Get all the set camera configurations
 * @param ptr Pointer to class
 * @param cam_configs Multiple camera configurations
 * @param capacity The amount of data the above pointer can hold
 * @param error Pointer to error
 * @return Number of pointers filled or complete array size
 */
TOF_EXPORT size_t tof_camera_get_multiple_camera_configurations(tof_camera_ct ptr, tof_camera_config_t* cam_configs, size_t capacity, tof_error_t *error);

/** Get all the set processing configurations
 * @param ptr Pointer to class
 * @param processing_configs Multiple processing configurations
 * @param capacity The amount of data the above pointer can hold
 * @param error Pointer to error
 * @return Number of pointers filled or complete array size
 */
TOF_EXPORT size_t tof_camera_get_multiple_processing_configurations(tof_camera_ct ptr, tof_processing_config_t* processing_configs, size_t capacity, tof_error_t *error);

/** Get the name for each configuration
 * @param ptr Pointer to class
 * @param names Names
 * @param capacity The amount of data the above pointer can hold
 * @param error Pointer to error
 * @return Number of pointers filled or complete array size
 */
TOF_EXPORT size_t tof_camera_get_multiple_names(tof_camera_ct ptr, const char** names, size_t capacity, tof_error_t *error);

/** Get the description for each configuration
 * @param ptr Pointer to class
 * @param description Descriptions
 * @param capacity The amount of data the above pointer can hold
 * @param error Pointer to error
 * @return Number of pointers filled or complete array size
 */
TOF_EXPORT size_t tof_camera_get_multiple_descriptions(tof_camera_ct ptr, const char** description, size_t capacity, tof_error_t *error);

/** A persistent ip data structure
 */
typedef struct tof_persistent_ip* tof_persistent_ip_t;

typedef struct tof_persistent_ip const* tof_persistent_ip_ct;

/** Destruct tof_persistent_ip */
TOF_EXPORT void tof_persistent_ip_delete(tof_persistent_ip_t ptr);

/** Create persistent ip
 * @param ptr Pointer to class
 * @param ip IP address
 * @param netmask Netmask
 * @param gateway Gateway
 * @param enabled Enable or disable persistent ip
 * @param error Pointer to error
 * @return New object
 */
TOF_EXPORT tof_persistent_ip_t tof_persistent_ip_new(const char* ip, const char* netmask, const char* gateway, bool enabled, tof_error_t *error);

/** Get the IP address
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return the IP address
 */
TOF_EXPORT const char* tof_persistent_ip_get_ip(tof_persistent_ip_ct ptr, tof_error_t *error);

/** Set the IP address
 * @param ptr Pointer to class
 * @param ip the IP address
 * @param error Pointer to error
 */
TOF_EXPORT void tof_persistent_ip_set_ip(tof_persistent_ip_t ptr, const char* ip, tof_error_t *error);

/** Get the netmask
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The netmask
 */
TOF_EXPORT const char* tof_persistent_ip_get_netmask(tof_persistent_ip_ct ptr, tof_error_t *error);

/** Set the netmask
 * @param ptr Pointer to class
 * @param netmask The netmask
 * @param error Pointer to error
 */
TOF_EXPORT void tof_persistent_ip_set_netmask(tof_persistent_ip_t ptr, const char* netmask, tof_error_t *error);

/** Get the gateway
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The gateway
 */
TOF_EXPORT const char* tof_persistent_ip_get_gateway(tof_persistent_ip_ct ptr, tof_error_t *error);

/** Set the gateway
 * @param ptr Pointer to class
 * @param gateway The gateway
 * @param error Pointer to error
 */
TOF_EXPORT void tof_persistent_ip_set_gateway(tof_persistent_ip_t ptr, const char* gateway, tof_error_t *error);

/** Get if persistent ip is enabled
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Is enabled
 */
TOF_EXPORT bool tof_persistent_ip_get_enabled(tof_persistent_ip_ct ptr, tof_error_t *error);

/** Set if persistent ip is enabled
 * @param ptr Pointer to class
 * @param enabled Is enabled
 * @param error Pointer to error
 */
TOF_EXPORT void tof_persistent_ip_set_enabled(tof_persistent_ip_t ptr, bool enabled, tof_error_t *error);

/** Functions shared across all live cameras
 */
typedef struct tof_live_camera* tof_live_camera_t;

typedef struct tof_live_camera const* tof_live_camera_ct;

/** Destruct tof_live_camera */
TOF_EXPORT void tof_live_camera_delete(tof_live_camera_t ptr);

/** Get the tof library version running on the camera
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Version
 */
TOF_EXPORT const char* tof_live_camera_version(tof_live_camera_t ptr, tof_error_t *error);

/** Check whether the camera is capable of updating remotely
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Remote update capable
 */
TOF_EXPORT bool tof_live_camera_remote_update_capable(tof_live_camera_t ptr, tof_error_t *error);

/** Update the camera remotely. After uploading the update file the function
 * will return. If a correct update file was uploaded, the camera will flash the
 * update and reboot. This generally takes about a minute. The camera can also
 * be updated via a web interface by visiting the ip address of the camera on
 * port 8080, for example http://192.168.1.208:8080
 * @param ptr Pointer to class
 * @param file_location Location of the update file
 * @param error Pointer to error
 */
TOF_EXPORT void tof_live_camera_remote_update(tof_live_camera_t ptr, const char* file_location, tof_error_t *error);

/** Whether the camera is capable of setting persistent ip
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Persistent ip capable
 */
TOF_EXPORT bool tof_live_camera_persistent_ip_capable(tof_live_camera_ct ptr, tof_error_t *error);

/** Get the current persistent ip settings
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Persistent IP
 */
TOF_EXPORT tof_persistent_ip_t tof_live_camera_get_persistent_ip(tof_live_camera_t ptr, tof_error_t *error);

/** Set the persistent ip settings. To activate persistent ip settings run
 * apply_persistent_ip. 
 * @param ptr Pointer to class
 * @param persistent_ip The persistent ip
 * @param error Pointer to error
 */
TOF_EXPORT void tof_live_camera_set_persistent_ip(tof_live_camera_t ptr, tof_persistent_ip_ct persistent_ip, tof_error_t *error);

/** Whether the camera is capable of applying the persistent ip
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Capable
 */
TOF_EXPORT bool tof_live_camera_apply_persistent_ip_capable(tof_live_camera_ct ptr, tof_error_t *error);

/** Apply persistent ip by rebooting the camera, power cycling the camera is
 * discouraged because the applied ip settings might not stick.
 * @param ptr Pointer to class
 * @param error Pointer to error
 */
TOF_EXPORT void tof_live_camera_apply_persistent_ip(tof_live_camera_t ptr, tof_error_t *error);

/** Check whether the camera is capable of software trigger
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Software trigger capable
 */
TOF_EXPORT bool tof_live_camera_software_trigger_capable(tof_live_camera_t ptr, tof_error_t *error);

/** Software trigger the camera
 * @param ptr Pointer to class
 * @param error Pointer to error
 */
TOF_EXPORT void tof_live_camera_software_trigger(tof_live_camera_t ptr, tof_error_t *error);

/** Get delay between network packets
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The delay, depending on the camera software it is in nanoseconds or
 * in cpu instruction
 */
TOF_EXPORT uint32_t tof_live_camera_get_delay(tof_live_camera_ct ptr, tof_error_t *error);

/** Set delay between network packets
 * @param ptr Pointer to class
 * @param delay The delay, depending on the camera software it is in nanoseconds
 * or in cpu instruction
 * @param error Pointer to error
 */
TOF_EXPORT void tof_live_camera_set_delay(tof_live_camera_t ptr, uint32_t delay, tof_error_t *error);

/** Get maximum size of each network packet transmitted
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Packet size in bytes
 */
TOF_EXPORT uint16_t tof_live_camera_get_packet_size(tof_live_camera_ct ptr, tof_error_t *error);

/** Set maximum size of each network packet transmitted
 * @param ptr Pointer to class
 * @param size Packet size in bytes
 * @param error Pointer to error
 */
TOF_EXPORT void tof_live_camera_set_packet_size(tof_live_camera_t ptr, uint16_t size, tof_error_t *error);

/** Whether the camera is capable of returning temperatures
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Capable
 */
TOF_EXPORT bool tof_live_camera_get_temperatures_capable(tof_live_camera_t ptr, tof_error_t *error);

/** Get the current on camera temperatures
 * @param ptr Pointer to class
 * @param temperatures Temperatures
 * @param capacity The amount of data the above pointer can hold
 * @param error Pointer to error
 * @return Number of pointers filled or complete array size
 */
TOF_EXPORT size_t tof_live_camera_get_temperatures(tof_live_camera_t ptr, float* temperatures, size_t capacity, tof_error_t *error);

/** Get voltage level on given pin
 * @param ptr Pointer to class
 * @param pin Pin
 * @param error Pointer to error
 * @return Voltage 3.3V
 */
TOF_EXPORT bool tof_live_camera_get_gpio(tof_live_camera_t ptr, size_t pin, tof_error_t *error);

/** Set voltage level on given pin
 * @param ptr Pointer to class
 * @param pin Pin
 * @param level Voltage level
 * @param error Pointer to error
 */
TOF_EXPORT void tof_live_camera_set_gpio(tof_live_camera_t ptr, size_t pin, bool level, tof_error_t *error);

#ifdef __cplusplus
}
#endif

#endif
