#ifndef _CHRONOPTICS_TOF_USER_CONFIG_HPP_
#define _CHRONOPTICS_TOF_USER_CONFIG_HPP_

#include <chronoptics/tof/user_config.h>

#include <chronoptics/tof/camera_config.hpp>
#include <chronoptics/tof/camera.hpp>
#include <chronoptics/tof/processing_config.hpp>

namespace chronoptics {
namespace tof {

/** Optimized configurations for different use cases
 */
enum class ConfigMode {
  LONG_RANGE_DYNAMIC_SCENE = 1, /** Suitable for most applications */
  MEDIUM_RANGE_DYNAMIC_SCENE = 2, /** For fast moving applications */
  MEDIUM_RANGE_BALANCED = 3, /** Balanced between range, speed and accuracy */
  SHORT_RANGE_HIGH_SPEED = 4, /** Maximum speed mode for short range and fast moving applications */
  LONG_RANGE_HIGH_RESOLUTION = 5, /** Full resolution long range */
  MEDIUM_RANGE_STATIC_SCENE = 6, /** Full resolution for slow changing scene */
  FAR_OUT_STATIC_SCENE = 7, /** Full resolution for slow changing scene */
};

/** Get the camera config associated with the config mode
 * @param camera Camera
 * @param config_mode Config mode
 * @return Camera Config
 */
inline CameraConfig config_mode_camera_config(Camera &camera, ConfigMode config_mode) {
  CameraConfig new_camera_config(static_cast<tof_camera_config_t>(nullptr));
  auto ptr = reinterpret_cast<tof_camera_config_t*>(&new_camera_config);
  *ptr = tof_config_mode_camera_config(*reinterpret_cast<tof_camera_t*>(&camera), static_cast<tof_config_mode>(config_mode), TOF_ERROR_HANDLER{});
  return new_camera_config;
}

/** Get the processing config associated with the config mode
 * @param camera Camera
 * @param config_mode Config mode
 * @return Processing Config
 */
inline ProcessingConfig config_mode_processing_config(Camera &camera, ConfigMode config_mode) {
  ProcessingConfig new_processing_config(static_cast<tof_processing_config_t>(nullptr));
  auto ptr = reinterpret_cast<tof_processing_config_t*>(&new_processing_config);
  *ptr = tof_config_mode_processing_config(*reinterpret_cast<tof_camera_t*>(&camera), static_cast<tof_config_mode>(config_mode), TOF_ERROR_HANDLER{});
  return new_processing_config;
}

/** Check whether hdr is possible with current camera config
 * @param config The camera config to check
 * @return Whether hdr is possible
 */
inline bool hdr_possible(const CameraConfig &config) {
  return tof_hdr_possible(*reinterpret_cast<const tof_camera_config_t*>(&config), TOF_ERROR_HANDLER{});
}

/** Check whether phase unwrap is possible with current camera config
 * @param config The camera config to check
 * @return Whether phase unwrap is possible
 */
inline bool phase_unwrap_possible(const CameraConfig &config) {
  return tof_phase_unwrap_possible(*reinterpret_cast<const tof_camera_config_t*>(&config), TOF_ERROR_HANDLER{});
}

/** Generate processing config based on camera configuration and calibration.
 * @param camera A camera
 */
inline void apply_default_processing_config(Camera &camera) {
  return tof_apply_default_processing_config(*reinterpret_cast<tof_camera_t*>(&camera), TOF_ERROR_HANDLER{});
}

} // tof
} // chronoptics

#endif
