import ctypes
from .__library__ import lib, get_error
from .camera import *
from .processing_config import *
from .camera_config import *


class KeaCamera(LiveCamera):
    """
    The main interface to the kea camera
    """
    def __del__(self):
        if self._ptr:
            lib.tof_kea_camera_delete(self._ptr)
            self._ptr = ctypes.c_void_p()


    def __init__(self, serial = None, processing_config = None, camera_config = None, ptr = None):
        """
        Construct new kea_camera class
        ----------

        Construct the kea camera with the given processing and camera config.

        Parameters
        ----------
        serial : str
                The serial number of the camera to connect to, if the camera is
                available over both usb and gige it will be connected over usb. If
                an empty string is supplied the camera will connect to the first usb
                camera, if no usb camera is connected it will connect to the first
                gige camera that replies (this reply time is random, so you might
                connect to a different camera each time if you've got multiple
                cameras connected to your network). If no camera is found with the
                given serial number a nullptr and error is returned.
        processing_config : ProcessingConfig
                The processing config
        camera_config : CameraConfig
                The camera config
        ----------

        Construct the kea camera

        Parameters
        ----------
        serial : str
                The serial number of the camera to connect to, if the camera is
                available over both usb and gige it will be connected over usb. If
                an empty string is supplied the camera will connect to the first usb
                camera, if no usb camera is connected it will connect to the first
                gige camera that replies (this reply time is random, so you might
                connect to a different camera each time if you've got multiple
                cameras connected to your network). If no camera is found with the
                given serial number a nullptr and error is returned.
        processing_config : ProcessingConfig
                The processing config
        ----------

        Construct the kea camera with a processing config generated by
        default_processing from the camera config

        Parameters
        ----------
        serial : str
                The serial number of the camera to connect to, if the camera is
                available over both usb and gige it will be connected over usb. If
                an empty string is supplied the camera will connect to the first usb
                camera, if no usb camera is connected it will connect to the first
                gige camera that replies (this reply time is random, so you might
                connect to a different camera each time if you've got multiple
                cameras connected to your network). If no camera is found with the
                given serial number a nullptr and error is returned.
        """
        self._ptr = ctypes.c_void_p()

        if ptr != None:
            self._ptr = ptr
        elif serial != None and processing_config != None and camera_config != None:
            self._ptr = lib.tof_kea_camera_new_both_configs(serial.encode("utf-8"), processing_config._ptr, camera_config._ptr, get_error())
        elif serial != None and processing_config != None:
            self._ptr = lib.tof_kea_camera_new(serial.encode("utf-8"), processing_config._ptr, get_error())
        elif serial != None:
            self._ptr = lib.tof_kea_camera_new_simple(serial.encode("utf-8"), get_error())
        else:
            raise RuntimeError("Can't construct empty class")

    def onCameraProcessingCapable(self):
        """
        Check whether this kea camera can do on camera processing

        Returns
        ----------
        bool
            On camera processing capable
        """
        return lib.tof_kea_camera_on_camera_processing_capable(self._ptr, get_error())


    def getOnCameraProcessing(self):
        """
        Get on camera processing

        Returns
        ----------
        bool
            True on camera processing, false on host processing
        """
        return lib.tof_kea_camera_get_on_camera_processing(self._ptr, get_error())


    def setOnCameraProcessing(self, on_camera_processing):
        """
        Set on camera processing

        Parameters
        ----------
        on_camera_processing : bool
            True on camera processing, false on host processing
        """
        return lib.tof_kea_camera_set_on_camera_processing(self._ptr, on_camera_processing, get_error())


    def setCameraConfig(self, camera_config):
        """
        Set the camera config

        Parameters
        ----------
        camera_config : CameraConfig
            Camera config
        """
        return lib.tof_kea_camera_set_camera_config(self._ptr, camera_config._ptr, get_error())


    def setConfigurations(self, camera_config, processing_config):
        """
        Set both the camera and processing config

        Parameters
        ----------
        camera_config : CameraConfig
            Camera config
        processing_config : ProcessingConfig
            Processing config
        """
        return lib.tof_kea_camera_set_configurations(self._ptr, camera_config._ptr, processing_config._ptr, get_error())


    def multipleConfigurationsPossible(self):
        """
        Check whether camera supports multiple configurations

        Returns
        ----------
        bool
            Multiple configurations possible
        """
        return lib.tof_kea_camera_multiple_configurations_possible(self._ptr, get_error())


    def setMultipleConfigurations(self, camera_configs, pro_configs):
        """
        Set multiple configurations

        Parameters
        ----------
        camera_configs : List[CameraConfig]
            Camera configurations
        pro_configs : List[ProcessingConfig]
            Processing configurations
        """
        return lib.tof_kea_camera_set_multiple_configurations(self._ptr, (ctypes.c_void_p * len(camera_configs))(*[item._ptr for item in camera_configs]), len(camera_configs), (ctypes.c_void_p * len(pro_configs))(*[item._ptr for item in pro_configs]), len(pro_configs), get_error())


    def setProcessConfig(self, config):
        """
        the processing config of the camera

        Parameters
        ----------
        config : ProcessingConfig
            The processing config
        """
        return lib.tof_kea_camera_set_process_config(self._ptr, config._ptr, get_error())


class DiscoveredCamera():
    """
    Simple class containing information about found cameras
    """
    def __del__(self):
        if self._ptr:
            lib.tof_discovered_camera_delete(self._ptr)
            self._ptr = ctypes.c_void_p()


    def __init__(self, ptr = None):
        """
        Construct new discovered_camera class
        """
        self._ptr = ctypes.c_void_p()

        if ptr != None:
            self._ptr = ptr
        else:
            raise RuntimeError("Can't construct empty class")

    def serial(self):
        """
        The serial

        Returns
        ----------
        str
            Serial
        """
        return lib.tof_discovered_camera_serial(self._ptr, get_error()).decode("utf-8")


    def info(self):
        """
        Additional information about the discovered camera

        Returns
        ----------
        str
            Additional information
        """
        return lib.tof_discovered_camera_info(self._ptr, get_error()).decode("utf-8")


    def model(self):
        """
        Get the camera model name

        Returns
        ----------
        str
            Camera model
        """
        return lib.tof_discovered_camera_model(self._ptr, get_error()).decode("utf-8")


    def isTui(self):
        """
        Check whether this camera is a TUI camera

        Returns
        ----------
        bool
            Is tui
        """
        return lib.tof_discovered_camera_is_tui(self._ptr, get_error())


def discoverCameras():
    """
    Discover all available cameras

    Returns
    ----------
    List[DiscoveredCamera]
        All available cameras
    """
    size = lib.tof_discover_cameras(ctypes.c_void_p(), 0, get_error())
    array = (ctypes.c_void_p * size)()
    lib.tof_discover_cameras(array, size, get_error())
    return [DiscoveredCamera(ptr = item) for item in array]

def createKeaCameraGige(processing_config, serial, packet_size = 1472):
    """
    Construct a kea camera using the gige interface. This could be useful in two
    cases. The first is when the camera is connected with both usb and ethernet but
    you want to use ethernet. The second is when for some reason the default maximum
    packet size is too large.

    Parameters
    ----------
    processing_config : ProcessingConfig
        The processing config
    serial : str
        The serial number of the camera to connect to, if the camera is available
        over both usb and gige it will be connected over usb. If an empty string is
        supplied the camera will connect to the first usb camera, if no usb camera
        is connected it will connect to the first gige camera that replies (this
        reply time is random, so you might connect to a different camera each time
        if you've got multiple cameras connected to your network). If no camera is
        found with the given serial number a nullptr and error is returned.
    packet_size : int
        The packet size to use for communication over ethernet

    Returns
    ----------
    KeaCamera
        A kea camera
    """
    return KeaCamera(ptr = lib.tof_create_kea_camera_gige(processing_config._ptr, serial.encode("utf-8"), packet_size, get_error()))


